// Responsible to download a file each interval, and provides a callback event to handle the results

const http = require('http'); 
const fs = require('fs');

const tracked_files = new Map();

zeek.on('Corelight_Http_Downloader::track_http_file', (config) => {
    let file_etag = '"None"';

    const interval_id = setInterval(() => {
        const http_options = {
            headers: { "If-None-Match": file_etag }
        };

        const request = http.get(config.src, http_options, function(response) {
            let is_changed = true;

            if(response.statusCode == 200) {
                const file = fs.createWriteStream(config.dest);
                response.pipe(file);

                // after download completes, close stream and alert zeek
                file.on("finish", () => {
                    file.close(() => {
                        zeek.invoke("Reporter::info", [`Intel file ${config.src} downloaded to ${config.dest}`])
                        zeek.event("Corelight_Http_Downloader::http_file_download_complete", [config.src, config.dest, is_changed]);
                        
                        // Update etag for next interval
                        if(response.headers.hasOwnProperty("etag")){
                            file_etag = response.headers["etag"];
                        }
                    });
                });
                return;
            }

            // For everything else, discard any data to avoid a memory leak
            response.resume();

            // If 304, file with etag was already downloaded. Skip both write and update
            if(response.statusCode == 304) {
                is_changed = false;
                zeek.event("Corelight_Http_Downloader::http_file_download_complete", [config.src, config.dest, is_changed]);
            }

            response.on('error', function (err) {
                zeek.invoke("Reporter::info", [`[response] Unable to download intel file ${config.src}: ${err}`])
            });
        });

        request.on('error', function (err) {
            zeek.invoke("Reporter::info", [`[request] Unable to download intel file ${config.src}: ${err}`])
        });
    }, parseInt(config.interval) * 1000);

    if(tracked_files.has(config.src)) {
        // Already tracking URL, so delete old interval
        clearInterval(tracked_files.get(config.src));
        tracked_files.delete(src);
    }
    tracked_files.set(config.src, interval_id);
});

zeek.on('Corelight_Http_Downloader::halt_track_http_file', (src) => {
    if(tracked_files.has(src)) {
        clearInterval(tracked_files.get(src));
        tracked_files.delete(src);
    }
});