module AMQP;

export {

    redef enum Log::ID += { AMQP_LOG };

    type Info: record {
		## Time
		ts: time &log &default=network_time();

		## Unique ID for the connection
		uid: string &log;

		## The connection's 4-tuple of endpoint addresses/ports
		id: conn_id &log;

        ## Is orig info
        is_orig: bool &log;

        ## AMQP specific fields
        ver_maj: count &log &optional;
        ver_min: count &log &optional;
        ver_rev: count &log &optional;
        host: string &log &optional;
        name: string &log &optional;
        product: string &log &optional;
        version: string &log &optional;
        copyright: string &log &optional;
        platform: string &log &optional;
        information: string &log &optional;
        instance: string &log &optional;

        ## Exchanges seen in basic.publish messages
        delivered_exchanges: set[string] &log &optional;

        ## Exchanges seen in basic.deliver messages
        published_exchanges: set[string] &log &optional;

        ## Queues seen in basic.consume messages
        consumed_queues: set[string] &log &optional;

        # Set to true to log, used internally
        log_me: bool &default=F;
    };

    ### Events ###

    # Raised for every AMQP message
    global AMQP::message: event(c: connection, is_orig: bool, ver_maj: count, ver_min: count, ver_rev: count);

    # Raised for every AMQP TableValueLongString
    global AMQP::TableValueLongString: event(c: connection, is_orig: bool, name: string, value:string);

    # Raised for every AMQP basic publish
    global AMQP::basic_publish: event(c: connection, is_orig: bool, exchange: string, routingkey: string);

    # Raised for every AMQP basic deliver
    global AMQP::basic_deliver: event(c: connection, is_orig: bool, consumer_tag: string, delivery_tag: count, exchange: string, routingkey: string);

    # Raised for every AMQP basic consume
    global AMQP::basic_consume: event(c: connection, is_orig: bool, consumer_tag: string, queue: string);

    # Log event
    global AMQP::log_amqp: event(rec: AMQP::Info);

}

redef record connection += {
	amqp_orig: Info &optional;
	amqp_resp: Info &optional;
};

function set_session(c: connection)
    {
    if ( ! c?$amqp_orig )
        {
        c$amqp_orig = [$uid=c$uid, $id=c$id, $is_orig=T];
        c$amqp_orig$delivered_exchanges = set();
        c$amqp_orig$published_exchanges = set();
        c$amqp_orig$consumed_queues = set();
        }
    if ( ! c?$amqp_resp )
        {
        c$amqp_resp = [$uid=c$uid, $id=c$id, $is_orig=F];
        c$amqp_resp$delivered_exchanges = set();
        c$amqp_resp$published_exchanges = set();
        c$amqp_resp$consumed_queues = set();
        }
    }

event AMQP::message(c: connection, is_orig: bool, ver_maj: count, ver_min: count, ver_rev: count)
    {
    set_session(c);
    if (is_orig)
        {
        c$amqp_orig$ver_maj = ver_maj;
        c$amqp_orig$ver_min = ver_min;
        c$amqp_orig$ver_rev = ver_rev;
        c$amqp_orig$log_me = T;
        }
    else
        {
        c$amqp_resp$ver_maj = ver_maj;
        c$amqp_resp$ver_min = ver_min;
        c$amqp_resp$ver_rev = ver_rev;        
        c$amqp_resp$log_me = T;
        }
    }

event AMQP::TableValueLongString(c: connection, is_orig: bool, name: string, value:string)
    {
    set_session(c);
    if (is_orig)
        {
        switch name 
            {
            case "information":
                c$amqp_orig$information=value;
                c$amqp_orig$log_me = T;
                break;
            case "instance":
                c$amqp_orig$instance=value;
                c$amqp_orig$log_me = T;
                break;
            case "version":
                c$amqp_orig$version=value;
                c$amqp_orig$log_me = T;
                break;
            case "copyright":
                c$amqp_orig$copyright=value;
                c$amqp_orig$log_me = T;
                break;
            case "platform":
                c$amqp_orig$platform=value;
                c$amqp_orig$log_me = T;
                break;
            case "name":
                c$amqp_orig$name=value;
                c$amqp_orig$log_me = T;
                break;
            case "host":
                c$amqp_orig$host=value;
                c$amqp_orig$log_me = T;
                break;
            case "product":
                c$amqp_orig$product=value;
                c$amqp_orig$log_me = T;
                break;
            }        
        }
    else
        {
        switch name 
            {
            case "information":
                c$amqp_resp$information=value;
                c$amqp_resp$log_me = T;
                break;
            case "instance":
                c$amqp_resp$instance=value;
                c$amqp_resp$log_me = T;
                break;
            case "version":
                c$amqp_resp$version=value;
                c$amqp_resp$log_me = T;
                break;
            case "copyright":
                c$amqp_resp$copyright=value;
                c$amqp_resp$log_me = T;
                break;
            case "platform":
                c$amqp_resp$platform=value;
                c$amqp_resp$log_me = T;
                break;
            case "name":
                c$amqp_resp$name=value;
                c$amqp_resp$log_me = T;
                break;
            case "host":
                c$amqp_resp$host=value;
                c$amqp_resp$log_me = T;
                break;
            case "product":
                c$amqp_resp$product=value;
                c$amqp_resp$log_me = T;
                break;
            }        
        }
    }

event AMQP::basic_publish(c: connection, is_orig: bool, exchange: string, routingkey: string)
    {
    set_session(c);
    local exchange_str: string = cat("x: ", exchange," rk: ", routingkey);
    if (is_orig)
        {
        add c$amqp_orig$published_exchanges[exchange_str];
        c$amqp_orig$log_me = T;
        }
    else
        {
        add c$amqp_resp$published_exchanges[exchange_str];
        c$amqp_resp$log_me = T;
        }
    }

event AMQP::basic_deliver(c: connection, is_orig: bool, consumer_tag: string, delivery_tag: count, exchange: string, routingkey: string)
    {
    set_session(c);
    local exchange_str: string = cat("x: ", exchange," rk: ", routingkey);
    if (is_orig)
        {
        add c$amqp_orig$delivered_exchanges[exchange_str];
        c$amqp_orig$log_me = T;
        }
    else
        {
        add c$amqp_resp$delivered_exchanges[exchange_str];
        c$amqp_resp$log_me = T;
        }
    }

event AMQP::basic_consume(c: connection, is_orig: bool, consumer_tag: string, queue: string)
    {
    set_session(c);
    if (is_orig)
        {
        add c$amqp_orig$consumed_queues[queue];
        c$amqp_orig$log_me = T;
        }
    else
        {
        add c$amqp_resp$consumed_queues[queue];
        c$amqp_resp$log_me = T;
        }
    }

event connection_state_remove(c: connection)
    {
    if (c?$amqp_orig && c$amqp_orig$log_me)
        Log::write(AMQP::AMQP_LOG, c$amqp_orig);
    if (c?$amqp_resp && c$amqp_resp$log_me)
        Log::write(AMQP::AMQP_LOG, c$amqp_resp);
    }

event zeek_init() &priority=5 
    {
    Log::create_stream(AMQP::AMQP_LOG, [$columns=Info, $ev=log_amqp, $path="amqp"]);
    }