module Telnet;

export {
	## Log stream identifier.
	redef enum Log::ID += { LOG };

	global log_policy: Log::PolicyHook;

	## How many bytes, minimum, of the connection to include in the logs.
	option logged_bytes = 2048;

	## Record type containing the column fields of the Telnet log.
	type Info: record {
		## Timestamp for when the activity happened.
		ts: time &log;
		## Unique ID for the connection.
		uid: string &log;
		## The connection's 4-tuple of endpoint addresses/ports.
		id: conn_id &log;
		## If this Telnet session is also TN3270.
		tn3270: bool &optional &log;
		## If this Telnet session is also TN3270E.
		tn3270e: bool &optional &log;
		## Terminal Type.
		terminal_type: string &optional &log;
		## Terminal Speed.
		terminal_speed: string &optional &log;
		## X Display Location.
		x_display_location: string &optional &log;
		## Names to environment variables, in order.
		environ_value_names: vector of string &optional &log;
		## Values to environment variables, in order.
		environ_value_values: vector of string &optional &log;
		## The device type provided in a device type request.
		tn3270e_device_type_request: string &optional &log;
		## The associate value provided in a device type request.
		tn3270e_device_type_request_associate: string &optional &log;
		## The connect value provided in a device type request.
		tn3270e_device_type_request_connect: string &optional &log;
		## The device type reply to a request.
		tn3270e_device_type_is: string &optional &log;
		## The associate value provided in a device type reply to a request.
		tn3270e_device_type_is_associate: string &optional &log;
		## The connect value provided in a device type reply to a request.
		tn3270e_device_type_is_connect: string &optional &log;
		## The first few bytes from the connection.
		data: string &optional &log;
	};

	type TN3270Attribute: record {
		attrib_type: Telnet::TN3270AttributeType;
		attrib_value: count;
	};

	## A function to translate EBCDIC strings to an ASCII representation.
	global EBCDIC_to_ASCII: function(e: string): string;

	## Default hook into Telnet logging.
	global log_telnet: event(rec: Info);

	# Telnet

	## Raised for telnet control codes, except for suboptions.
	global control_code: event(c: connection, is_orig: bool, code: Telnet::TelnetCode, option_code: Telnet::TelnetOptionCode);
	## Raised for telnet suboptions.  Payload is provided if the type is not fully parsed by Spicy elsewhere.
	global sub_option: event(c: connection, is_orig: bool, subcommand: Telnet::TelnetOptionCode, payload: string);
	## Raised for telnet terminal types.
	global terminal_type: event(c: connection, is_orig: bool, ttype: string);
	## Raised for terminal speed.  
	global terminal_speed: event(c: connection, is_orig: bool, tspeed: string);
	## Raised for X display location.
	global x_display_location: event(c: connection, is_orig: bool, location: string);
	## Raised for each telnet environment variable.
	global environ_value: event(c: connection, is_orig: bool, env_type: Telnet::NewEvironTypes, env_value: string);
	## Raised for all telnet data.  Not raised if the connection is TN3270.
	global data: event(c: connection, is_orig: bool, data: string);

	# TN3270
	# Note that TN3270 data is in EBCDIC, not ASCII.

	## Raised for each TN3270 orig message.
	global tn3270_orig: event(c: connection, is_orig: bool, aid: Telnet::TN3270AID);
	## Raised for each TN3270 resp message.
	global tn3270_resp: event(c: connection, is_orig: bool, command_code: Telnet::TN3270CommandCode);
	## Raised for every TN3270 order.
	global tn3270_order: event(c: connection, is_orig: bool, order_type: Telnet::TN3270OrderType);
	## Raised for every TN3270 SF order.
	global tn3270_order_sf: event(c: connection, is_orig: bool, attrib: count);
	## Raised for every TN3270 SFE order.
	global tn3270_order_sfe: event(c: connection, is_orig: bool, attributes: vector of TN3270Attribute);
	## Raised for every TN3270 SA order.
	global tn3270_order_sa: event(c: connection, is_orig: bool, attrib_type: Telnet::TN3270AttributeType, attrib_value: count);
	## Raised for every TN3270 SBA order.
	global tn3270_order_sba: event(c: connection, is_orig: bool, buffer_address: count);
	## Raised for every TN3270 MF order.
	global tn3270_order_mf: event(c: connection, is_orig: bool, attributes: vector of TN3270Attribute);
	## Raised for every TN3270 RA order.
	global tn3270_order_ra: event(c: connection, is_orig: bool, stop_address: count, char_to_be_repeated: count);
	## Raised for every TN3270 EUA order.
	global tn3270_order_eua: event(c: connection, is_orig: bool, stop_address: count);
	## Raised for every TN3270 GE order.
	global tn3270_order_ge: event(c: connection, is_orig: bool, character_code: count);
	## Raised for the data found between TN3270 orders.
	global tn3270_data: event(c: connection, is_orig: bool, data: string);

	# TN3270E

	## Raised for the data found in TN3270E headers.
	global tn3270e_header: event(c: connection, is_orig: bool, data_type: Telnet::TN3270EDataType, request_flag: count, response_flag: count, seq_number: count, payload: string);
	## Raised for the functions telnet TN3270 suboption.
	global telnet_tn3270e_suboption_functions: event(c: connection, is_orig: bool, functions: string);
	## Raised for the device type telnet TN3270 suboption.
	global telnet_tn3270e_suboption_device_type: event(c: connection, is_orig: bool, is_request: bool, device_type: string, connect: string, associate: string);
	## Raised for the device type reject reason.
	global telnet_tn3270e_reject_reason: event(c: connection, is_orig: bool, reason: count);

}

redef record connection += {
	telnet: Info &optional;
};

const ports = {
	23/tcp # adapt port number in telnet.evt accordingly
};

redef likely_server_ports += { ports };

event zeek_init() &priority=5
	{
	Log::create_stream(Telnet::LOG, [$columns=Info, $ev=log_telnet, $path="telnet", $policy=log_policy]);
	}

# Initialize logging state.
hook set_session(c: connection)
	{
	if ( c?$telnet )
		return;

	c$telnet = Info($ts=network_time(), $uid=c$uid, $id=c$id, $environ_value_names=vector(), $environ_value_values=vector());
	}

function emit_log(c: connection)
	{
	if ( ! c?$telnet )
		return;

	Log::write(Telnet::LOG, c$telnet);
	delete c$telnet;
	}

event connection_state_remove(c: connection) &priority=-5
	{
	emit_log(c);
	}

event terminal_type(c: connection, is_orig: bool, ttype: string) 
	{
	hook set_session(c);
	c$telnet$terminal_type = ttype;
	}

event terminal_speed(c: connection, is_orig: bool, tspeed: string) 
	{
	hook set_session(c);
	c$telnet$terminal_speed = tspeed;
	}

event x_display_location(c: connection, is_orig: bool, location: string)
	{
	hook set_session(c);
	c$telnet$x_display_location = location;
	}

event environ_value(c: connection, is_orig: bool, env_type: Telnet::NewEvironTypes, env_value: string)
	{
	hook set_session(c);

	if (env_type == Telnet::NewEvironTypes_Var)
		c$telnet$environ_value_names += env_value;

	if (env_type == Telnet::NewEvironTypes_ValueVal)
		c$telnet$environ_value_values += env_value;
	}

event tn3270_resp(c: connection, is_orig: bool, command_code: Telnet::TN3270CommandCode)
	{
	hook set_session(c);

	c$telnet$tn3270 = T;
	}

event tn3270_orig(c: connection, is_orig: bool, aid: Telnet::TN3270AID)
	{
	hook set_session(c);

	c$telnet$tn3270 = T;
	}

event tn3270e_header(c: connection, is_orig: bool, data_type: Telnet::TN3270EDataType, request_flag: count, response_flag: count, seq_number: count, payload: string)
	{
	hook set_session(c);

	c$telnet$tn3270 = T;
	c$telnet$tn3270e = T;
	}

event telnet_tn3270e_suboption_device_type(c: connection, is_orig: bool, is_request: bool, device_type: string, connect_val: string, associate_val: string)
    {
	hook set_session(c);

	c$telnet$tn3270 = T;
	c$telnet$tn3270e = T;

	if (is_request)
		{
		c$telnet$tn3270e_device_type_request = device_type;
		if (|associate_val| > 0)
			c$telnet$tn3270e_device_type_request_associate = associate_val;
		if (|connect_val| > 0)
			c$telnet$tn3270e_device_type_request_connect = connect_val;
		}
	else
		{
		c$telnet$tn3270e_device_type_is = device_type;
		if (|associate_val| > 0)
			c$telnet$tn3270e_device_type_is_associate = associate_val;
		if (|connect_val| > 0)
			c$telnet$tn3270e_device_type_is_connect = connect_val;
		}
    }

function EBCDIC_to_ASCII(e: string): string
	{
	local a = "";

	for ( ei in e )
		a += EBCDIC_mappings[string_to_ascii_hex(ei)];

	return a;
	}

# Adds the telnet/tn3270 content to the log.
function add_content_to_log(c: connection, data: string, is_ebcdic: bool)
	{
	hook set_session(c);

	if ( c$telnet?$data && |c$telnet$data| >= logged_bytes )
		return;
	
	if ( ! c$telnet?$data )
		c$telnet$data = "";

	if (is_ebcdic)
		c$telnet$data += EBCDIC_to_ASCII(data);
	else
		c$telnet$data += data;
	}

# Add connection content data to the log.
event data(c: connection, is_orig: bool, data: string)
	{
	add_content_to_log(c, data, F);
	}

# Add connection content data to the log.
event tn3270_data(c: connection, is_orig: bool, data: string) 
	{
	add_content_to_log(c, data, T);
	}